/*
 *
 *  dumpcheck - a simple command line utility to report illegal UTF-8
 *  sequences and illegal XML characters in the ODP data dumps.
 * 
 *  Version 1.11
 *
 *  History
 * 
 *  14 Feb 2003 v1.0 rsr  - Initial release
 *  15 Feb 2003 v1.1 rsr  - UTF-8 sequences with incorrect length were causing
 *                          subsequent contiguous sequences to be read 
 *                          incorrectly. Fixed.
 *  24 Feb 2003 v1.2 rsr  - Added total error count to output
 *                          Increased buffer size to 32Kb
 *                          Added code to look for catids. The code will fail
 *                          if a catid sequence crosses a buffer boundry but
 *                          it should identify the catid most of the time.
 *  25 Apr 2003 v1.3 rsr  - Added individual counters for UTF8 and XML errors
 *                          Added 2nd pass to check XML well-formedness
 *  16 Jul 2003 v1.4 rsr  - Normalized singular/plural output text
 *                          Added hex offset values
 *  06 Aug 2003 v1.5 rsr  - Added line number for XML/UTF-8 byte errors
 *                          Improved UTF-8 error detection logic
 *                          Added libxml2 error reporting for well-formedness
 *  21 Mar 2004 v1.6 rsr  - Reduced size of output by trimming error messages
 *                          Changed to name from rdfcheck to dumpcheck.
 *  05 Jun 2004 v1.7 rsr  - Added check for out of range Unicode values.
 *  05 Jul 2004 v1.8 rsr  - Added improved libxml2 line number check suggested
 *                          by William M. Brack <wbrack@mmm.com.hk>
 *  01 Oct 2004 v1.9 rsr  - Changes from Andreas Steinmetz <ast@domdv.de>
 *                          : Fix to CFLAGs in Makefile, added -O2 optimizing
 *                          : Added missing include for ctype.h
 *                          : Assorted printf formatting fixes
 *                          : dumpcheck now returns 1 on error 0 on success
 *                          : reads gzip compressed or uncompressed input
 *  29 Jun 2007 v1.10 rsr - Updated license to GPLv3
 *  30 Nov 2007 v1.11 rsr - Minor type casting to fix GCC 4.x warnings.
 *
 *------------------------------------------------------------------------
 * Copyright (c) 2003-2007 R. Steven Rainwater
 * Licensed under the GNU GPL version 3 or later
 *
 * Portions of this code are derived from the UTF-8 validation function
 * example code provided by Unicode, Inc.:
 *
 * Copyright 2001 Unicode, Inc.
 * 
 * Disclaimer
 * 
 * This source code is provided as is by Unicode, Inc. No claims are
 * made as to fitness for any particular purpose. No warranties of any
 * kind are expressed or implied. The recipient agrees to determine
 * applicability of information provided. If this file has been
 * purchased on magnetic or optical media from Unicode, Inc., the
 * sole remedy for any claim will be exchange of defective media
 * within 90 days of receipt.
 * 
 * Limitations on Rights to Redistribute This Code
 * 
 * Unicode, Inc. hereby grants the right to freely use the information
 * supplied in this file in the creation of products supporting the
 * Unicode Standard, and to make copies of this file in any form
 * for internal or external distribution as long as this notice
 * remains attached.
 * 
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <zlib.h>
#include <libxml/xmlreader.h>

#define TRUE 1
#define FALSE 0
#define BUFFSIZE 50000

static unsigned int isBadUnicode(unsigned char *, int);
static int isLegalUTF8(unsigned char *, int);
static int isLegalXML(unsigned char *);
static int isWellFormedXML(char *);

static const char UTF8length[256] = {
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5
};

int main(int argc, char *argv[])
{
    gzFile *f;
    int i = 0;
    int r = 0;
    int seq_length = 0;
    int seq_complete = TRUE;
    int WFerr = 0;
    int Xerr = 0;
    int Uerr = 0;
    int UCval = 0;
    unsigned long offset = 0;
    unsigned long line = 1;
    unsigned char seq[6];
    unsigned char *byte;
    unsigned char catid[8];
    unsigned char buff[BUFFSIZE];

    printf("\ndumpcheck version 1.11 (libxml2 v%s)\n",LIBXML_DOTTED_VERSION);

    if(argc!=2)
    {
        printf("usage: dumpcheck odpdump.rdf.u8\n");
        exit(1);
    }

    /* Pass 1, check for character encoding errors */

    if(!(f = gzopen(argv[1],"r"))) {
        fprintf(stderr,"Error: unable to open [%s]\n",argv[1]);
	exit(1);
    }

    /* While there's more file data to read...*/
    while(TRUE) {
	r = gzread(f, buff, BUFFSIZE);
	if(r==0) break;
	byte = &buff[0];

	/* While there are more bytes in this block of data...*/
	while(byte <= &buff[r-1]) {

            /* count lines */
	    if(*byte == '\n') { line++; }

	    /* look for a catid */
	    if(!strncmp((char *)byte,"<catid>",7)) {
	        for(i=0;(byte+7+i <= &buff[r-1]) && (isdigit(*(byte+7+i)));i++) { 
	          catid[i] = *(byte+7+i);
	        }
		if(*(byte+7+i) == '<') { catid[i] = 0; }
		else { memset(catid,0,sizeof(catid)); }
	    }

	    /* test for valid UTF-8 sequence */
	    if(seq_complete) {
		if(UTF8length[*byte] == 0) {
		    if(!isLegalUTF8(byte,1)) {
		      Uerr++;
		      printf("UTF-8 Error: Illegal sequence [0x%02x]\n\tOffset: %lu [0x%lx] Line: %lu, Cat ID: %s\n",*byte,offset,offset,line,catid);
		    }
		}
		else {
		    seq_complete = FALSE;
		    seq_length = 1;
		    seq[0] = *byte;
		}
	    }
	    else {
                /* append next byte to sequence */
		seq[seq_length] = *byte;
		seq_length++;
		if(UTF8length[seq[0]]+1 == seq_length || seq_length > 5) seq_complete = TRUE;
		if(seq_complete) {
		    if(!isLegalUTF8(seq,seq_length)) {
		        Uerr++;
		        printf("UTF-8 Error: Illegal sequence ");
			for(i=0;i<seq_length;i++) { printf(" [0x%02x]",seq[i]); }
		        printf("\n\tOffset: %lu [0x%lx] Line %lu, Cat ID: %s\n",offset,offset,line,catid);
		    }
		    else {
			if((UCval = isBadUnicode(seq,seq_length))) {
			    Uerr++;
			    printf("UTF-8 Error: Malformed Unicode value (U+%X) from sequence ", UCval);
			    for(i=0;i<seq_length;i++) { printf(" [0x%02x]",seq[i]); }
		    	    printf("\n\tOffset: %lu [0x%lx] Line %lu, Cat ID: %s\n",offset,offset,line,catid);
			}
		    }
		    memset(seq,0,6);
		}
		if(*byte < 0x80 && !isLegalUTF8(byte,1)) {
		    Uerr++;
	            printf("UTF-8 Error: Illegal sequence [0x%02x]\n\tOffset: %lu [0x%lx] Line: %lu, Cat ID: %s\n",*byte,offset,offset,line,catid);
		}
	    }
	
	    /* test for legal xml chars, ignoring incomplete UTF-8 sequences */
	    if(seq_complete && !isLegalXML(byte)) {
	        Xerr++;
		printf("XML Error: Illegal character [0x%02x]\n\tOffset: %lu [0x%lx] Line: %lu, Cat ID: %s\n",*byte,offset,offset,line,catid);
	    }

	    offset++;
	    byte++;
	}
    }
    
    gzclose(f);
    
    /* Pass 2, check for XML syntax errors using libxml2 */
    WFerr = isWellFormedXML(argv[1]);

    /* Report results */    
    printf("%lu bytes processed\nUTF-8 Sequence error(s): %i\nXML Character encoding error(s): %i\n",offset,Uerr,Xerr);
    if(WFerr) printf("W3C XML well-formedness: Passed\n");
    else printf("W3C XML well-formedness: Failed\n");
    if(WFerr || Uerr || Xerr) return 1;
    return 0;
}


/*
 * wfErrorFunc - used by isWellFormedXML to handle errors
 */
void wfErrorFunc(void *arg, const char *msg, xmlParserSeverities severity, xmlTextReaderLocatorPtr locator) {
  const char *err[5] = {
    "",
    "Validity Warning",
    "Validity Error",
    "Warning",
    "Error"
  };

  /* Determine line number */
  xmlParserCtxtPtr ctx = (xmlParserCtxtPtr)locator;
  int line;
  xmlParserInputPtr input;
  
  input = ctx->input;
  if ((input->filename == NULL) && (ctx->inputNr > 1))
    input = ctx->inputTab[ctx->inputNr - 2];
  if(input != NULL) {
    line = input->line;
  }
  else {
    line = -1;
  }
  
  printf("Well-formedness %s [line %i]: %s\n",err[severity],line,msg);
}


/* 
 * isWellFormedXML: Checks XML well-formedness acccording to the W3C 
 * XML 1.0 definition. http://www.w3.org/TR/REC-xml
 * Based on the streamFile function from xmllint.c, part of the libxml
 * library by Daniel Veillard.
 * http://xmlsoft.org/
 *
 */
static int isWellFormedXML(char *filename) {
    xmlTextReaderPtr reader;
    int ret = 1;

    reader = xmlNewTextReaderFilename(filename);
    if (reader != NULL) {
        xmlTextReaderSetErrorHandler(reader,wfErrorFunc,NULL);
	while (ret == 1) {
	    ret = xmlTextReaderRead(reader);
	}
	xmlFreeTextReader(reader);
    }
    else {
        fprintf(stderr,"Error: unable to open [%s]\n",filename);
	exit(1);
    }
    if(ret != 0) return 0;
    return 1;
}


/* 
 * isLegalXML: Validates character as a legal XML character 
 * based on the W3C XML 1.0 Recommendation:
 * http://www.w3.org/TR/1998/REC-xml-19980210
 *
 */
static int isLegalXML(unsigned char *c) {
    if(*c <= 0x08) return FALSE;
    if(*c == 0x0b) return FALSE;
    if(*c == 0x0c) return FALSE;
    if(*c >= 0x0e && *c <= 0x1f) return FALSE;
    return TRUE;
}


/*
 * isLegalUTF8:  Validation of a byte sequence as a legal
 * UTF-8 sequence. This function based on code by Unicode, Inc.
 *
 */
static int isLegalUTF8(unsigned char *source, int length) {
    unsigned char a;
    unsigned char *srcptr = source+length;
    switch (length) {
        default: return FALSE;
        /* Everything else falls through when "true"... */
        case 4: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return FALSE;
        case 3: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return FALSE;
        case 2: if ((a = (*--srcptr)) > 0xBF) return FALSE;
	        switch (*source) {
		    /* no fall-through in this inner switch */
		    case 0xE0: if (a < 0xA0) return FALSE; break;
		    case 0xF0: if (a < 0x90) return FALSE; break;
		    case 0xF4: if (a > 0x8F) return FALSE; break;
	            default:  if (a < 0x80) return FALSE;
	        }
        case 1: if (*source >= 0x80 && *source < 0xC2) return FALSE;
                if (*source > 0xF4) return FALSE;
	}
	return TRUE;
}


/*
 * isBadUnicode:  Decode the UTF-8 sequence and determine if the Unicode
 * character is legal. In particular we are checking for UTF-16 surrogates
 * U+D800 - U+DFFF and for the explicitly excluded characters U+FFFE and
 * U+FFFF which are not allowed in normal data. 
 *
 * This code is roughly based on code from the function xmlNextChar() in
 * Daniel Veillard's libxml2. 
 *
 */
static unsigned int isBadUnicode(unsigned char *source, int length) {
    unsigned int ucode = 0;
    if(*source & 0x80 && *source & 0xc0) {
	if(length == 4) {
	    ucode = (*source++ & 0x7) << 18;
	    ucode |= (*source++ & 0x3f) << 12;
	    ucode |= (*source++ & 0x3f) << 6;
	    ucode |= *source & 0x3f;
	}
        if(length == 3) {
	    ucode = (*source++ & 0xf) << 12;
	    ucode |= (*source++ & 0x3f) << 6;
	    ucode |= *source++ & 0x3f;
	}
	if(((ucode > 0xd7ff) && (ucode < 0xe000)) ||
	   ((ucode > 0xfffd) && (ucode < 0x10000)) ||
	   (ucode >= 0x110000)) {
	    return ucode;
	}
    }
    return 0;
}
